/*

  DEFCON China 1.0 Official Badge (2019): FPC to Arduino Shield

  Author: Joe Grand, Grand Idea Studio [@joegrand]

  Program Description:

  This program is used during DEFCON China 1.0 to communicate with the official badge.
  It is used as part of the conference game where attendees have to complete certain
  tasks in order to unlock LEDs on their badge.

  Depending on the DIP switch settings on the FPC to Arduino Shield hardware, a single
  root LED on the badge can be set or cleared. The status of all LEDs on the badge can
  also be read and passed to the host PC, which is used with the visual display at the
  event.

  Complete design documentation can be found at:
  http://www.grandideastudio.com/portfolio/defcon-china-2019-badge

*/

#include <SoftwareSerial.h>

/***************************************************************************
 ************************** Pin Assignments ********************************
 ***************************************************************************/

#define flexGPIO        9       // GPIO pin (via FPC connector), unused
#define flexTxPin       10      // Serial output (to badge's FPC)
#define flexRxPin       11      // Serial input (from badge's FPC)
#define avrSense        12      // Detect when badge is connected (when VIO from badge is supplied to the TXS0104), active HIGH

#define ledRed          A0      // Bi-color LED, red, active HIGH
#define ledGreen        A1      // Bi-color LED, green, active HIGH

#define swDIP1          7       // DIP switches
#define swDIP2          6
#define swDIP3          5
#define swDIP4          4
#define swDIP5          3
#define swDIP6          2

/***************************************************************************
 **************************** Definitions **********************************
 ***************************************************************************/

//#define DCN_TREE                // if this shield is being used for the DEFCON China tree/artwork display

#define DCN_UART_START  '<'     // start byte (sent from FPC to Arduino Shield)
#define DCN_UART_END    '>'     // end byte

#define DCN_BUF_SIZE    8       // maximum length of incoming data from the badge

#define TIME_DEBOUNCE       500     // time (in ms) badge needs to be inserted before continuing (debounce)
#define TIME_BADGE_CHECK    2000    // time (in ms) badge has to properly communicate
#define TIME_TRANSFER       2000    // maximum allowable time (in ms) of badge-to-shield data transfer (prevents hanging during aborted communications)

/****************************************************************************
 ************************** Global variables ********************************
 ***************************************************************************/

// Set up a new serial port for communication via FPC
SoftwareSerial flexSerial =  SoftwareSerial(flexRxPin, flexTxPin);

static byte b4 = 0;     // unique ID
static byte b5 = 0;
static byte b6 = 0;
static byte b7 = 0;
static byte b8 = 0;     // game flags (bit set to 1 when task is completed)
static byte b9 = 0;

// Arduino Serial Terminal
char receivedByte;
boolean newByte = false;

/****************************************************************************
 *************************** Constants **************************************
 ***************************************************************************/

/****************************************************************************
 ********************* Function Prototypes **********************************
 ***************************************************************************/

/****************************************************************************
 ************************** Functions ***************************************
 ***************************************************************************/

void setup()  // Setup code (called once on start-up)
{
    hardware_init();  // Low-level hardware initialization

    Serial.println("\n\rWelcome to the DEFCON China 1.0 FPC to Arduino Shield\n\r");
    Serial.flush();   // Wait for all bytes to be transmitted to the console
}

void hardware_init()
{
    // Define pin modes
    pinMode(flexTxPin, OUTPUT);
    pinMode(flexRxPin, INPUT);
    pinMode(avrSense, INPUT);
    pinMode(flexGPIO, INPUT);

    pinMode(ledRed, OUTPUT);
    pinMode(ledGreen, OUTPUT);

    pinMode(swDIP1, INPUT_PULLUP);
    pinMode(swDIP2, INPUT_PULLUP);
    pinMode(swDIP3, INPUT_PULLUP);
    pinMode(swDIP4, INPUT_PULLUP);
    pinMode(swDIP5, INPUT_PULLUP);
    pinMode(swDIP6, INPUT_PULLUP);

    setLEDOff();

    // Setup Arduino Serial Monitor
    Serial.begin(9600);
    while (!Serial);        // Wait until ready

    // Setup FPC Serial Port
    flexSerial.begin(9600);
    while (!flexSerial);    // Wait until ready
}

/**************************************************************/

void loop()  // Main code (runs repeatedly)
{
    int state, led;
      
    setLEDOrange();

    // Automatic badge detection
    Serial.println("[*] Waiting for badge");

    // Ensure badge is fully inserted into connector
    unsigned long previousLowTime = millis();   // Counter for timeout
    
    while (true) 
    {
        if (digitalRead(avrSense))  // If current state is HIGH, badge has been inserted
        { 
            if (((signed long)(millis() - previousLowTime)) > TIME_DEBOUNCE) // If badge has been inserted long enough, then proceed
                break;
        } 
        else 
        {
            previousLowTime = millis();
        }
    }

    if (digitalRead(avrSense))  // If badge is still connected, pin will still be HIGH
    {
        Serial.println("[*] Badge detected");
        setLEDRed();

        // Read the state of the DIP switches to determine function and desired LED
        state = 0;
        bitWrite(state, 1, !digitalRead(swDIP1));
        bitWrite(state, 0, !digitalRead(swDIP2));

        led = 0;    // Root LED select (0-15)
        bitWrite(led, 3, !digitalRead(swDIP3));
        bitWrite(led, 2, !digitalRead(swDIP4));
        bitWrite(led, 1, !digitalRead(swDIP5));
        bitWrite(led, 0, !digitalRead(swDIP6));

        // Now that we've detected the badge, try to communicate with it 
        // Once we achieve two identical reads, we know the badge is properly situated and ready to handle the desired command
        if (state != 0) 
        {
            unsigned long readTestStartTime = millis();     // counter for timeout
            char badgeDataPreviousCheck[DCN_BUF_SIZE];      // buffers to verify proper communication with badge
            char badgeDataCheck[DCN_BUF_SIZE];      
            memset(badgeDataPreviousCheck, 0xFF, DCN_BUF_SIZE);   // clear buffer
      
            while (((signed long)(millis() - previousLowTime)) < TIME_BADGE_CHECK) 
            {
                if (readBadgeData(badgeDataCheck)) 
                {
                    if (memcmp(badgeDataPreviousCheck, badgeDataCheck, DCN_BUF_SIZE) == 0)   // if both reads match, then proceed
                        break;
                    else 
                        memcpy(badgeDataPreviousCheck, badgeDataCheck, DCN_BUF_SIZE);   // store contents from first read into buffer
                }
            }
        }

        // data is expected by the badge in the following format: <command,value (optional)>
        // command = character or string
        // value = integer
        switch (state)
        {
            case 0:     // Off
                Serial.println("[*] System off");
                break;
            case 0b01:  // Set selected LED
                Serial.print("[*] Setting LED #");
                Serial.println(led, DEC);
                flexSerial.print("<S,");
                flexSerial.print(led, DEC);
                flexSerial.println(">");
                break;
            case 0b10:  // Clear selected LED
                Serial.print("[*] Clearing LED #");
                Serial.println(led, DEC);
                flexSerial.print("<C,");
                flexSerial.print(led, DEC);
                flexSerial.println(">");
                break;
            case 0b11:  // Read state of LEDs from badge
                Serial.println("[*] Reading LEDs from badge");
                char badgeData[DCN_BUF_SIZE];   // Buffer for incoming data
                boolean done = false;           // Bit set when data properly received
                done = readBadgeData(badgeData);

                if (done == true)
                {
                    // Move received data from our buffer to the individual global variables
                    b4 = badgeData[0];
                    b5 = badgeData[1];
                    b6 = badgeData[2];
                    b7 = badgeData[3];
                    b8 = badgeData[4];
                    b9 = badgeData[5];
                }
                else
                {
                    // Data transfer did not complete
                    b4 = b5 = b6 = b7 = b8 = b9 = 0xFF;
                }
#ifndef DCN_TREE
                if ((b4 == 0xFF) && (b5 == 0xFF) && (b6 == 0xFF) && (b7 == 0xFF))
                {
                    Serial.println("-> Error");
                }
                else
                {
                    Serial.print("-> Unique ID: 0x");
                    printHex(&b4, 1);
                    printHex(&b5, 1);
                    printHex(&b6, 1);
                    printHex(&b7, 1);
                    Serial.println();
        
                    Serial.print("-> Game flags: ");
                    printBits(b8);
                    Serial.print(" ");
                    printBits(b9);
                    Serial.print(" [0x");
                    printHex(&b8, 1);
                    printHex(&b9, 1);
                    Serial.println("]");
                }
#else   // If used for tree/artwork, send data in a simplified form
                Serial.print(DCN_UART_START);
                Serial.write(b4);   // unique ID
                Serial.write(b5);
                Serial.write(b6);
                Serial.write(b7);
                Serial.write(b8);   // game flags (bit set to 1 when task is completed)
                Serial.write(b9);
                Serial.println(DCN_UART_END);
#endif
                break;
        }
    }

    delay(500);
    setLEDGreen();
    Serial.println("[*] Done: Remove badge");

    while (digitalRead(avrSense));  // Wait until badge is removed
    Serial.println("[*] Badge removed");
    delay(1000);   // Delay before starting again
}

/**************************************************************/

boolean readBadgeData(uint8_t *buf) 
{
    while (flexSerial.read() >= 0); // Flush the receive buffer
    flexSerial.println("<R>");
  
    // Wait for a response from the badge
    char offset = 0;    // Offset into buffer
    unsigned long timeout = millis() + TIME_TRANSFER;  // Counter for timeout
    buf[0] = 0;     // Clear the buffer
  
    while (true)
    {
        if (flexSerial.available() > 0) // If there are any bytes available to read, then the badge has probably started responding
        {
            buf[offset] = flexSerial.read();        // Get the byte and store it in our buffer
            if (buf[offset] == DCN_UART_START)      // If we receive the start byte, then get ready to receive the actual data
            {
                offset = -1;     // Clear offset (will be incremented back to 0 at the end of the loop)
            }
            else if (buf[offset] == DCN_UART_END)   // If we receive the stop byte, then data transfer is complete
            {
                buf[offset] = 0;  // Null terminate the string of bytes we just received
                return true;
            }

            offset++;  // Increment offset into array
            if (offset >= DCN_BUF_SIZE) offset = 0; // If the incoming data string is longer than our buffer, wrap around to avoid going out-of-bounds
        }

        if ((int32_t)(millis() - timeout) > 0) break; // If data transmission has been interrupted (timed out), break out of the loop
    }

    return false;
}

/**************************************************************/

// print 8-bit data in hex with leading zero if needed
// from http://forum.arduino.cc/index.php?topic=38107#msg282343
void printHex(uint8_t *data, uint8_t length)
{
    char tmp[length * 2 + 1];
    byte first ;
    int j = 0;
  
    for (uint8_t i = 0; i < length; i++)
    {
        first = (data[i] >> 4) | 48;
        if (first > 57) 
            tmp[j] = first + (byte)39;
        else 
            tmp[j] = first ;
        j++;

        first = (data[i] & 0x0F) | 48;
        if (first > 57) 
            tmp[j] = first + (byte)39;
        else 
            tmp[j] = first;
        j++;
    }
  
    tmp[length * 2] = 0;
    Serial.print(tmp);
}

/**************************************************************/

// print all 8 bits of a byte including leading zeros
// from https://forum.arduino.cc/index.php?topic=46320.0
void printBits (byte myByte)
{
    for (byte mask = 0x80; mask; mask >>= 1)
    {
        if (mask  & myByte)
            Serial.print('1');
        else
            Serial.print('0');
    }
}

/**************************************************************/

void setLEDRed()
{
    digitalWrite(ledRed, HIGH);
    digitalWrite(ledGreen, LOW);
}

/**************************************************************/

void setLEDGreen()
{
    digitalWrite(ledRed, LOW);
    digitalWrite(ledGreen, HIGH);
}

/**************************************************************/

void setLEDOrange()
{
    digitalWrite(ledRed, HIGH);
    digitalWrite(ledGreen, HIGH);
}

/**************************************************************/

void setLEDOff()
{
    digitalWrite(ledRed, LOW);
    digitalWrite(ledGreen, LOW);
}

/**************************************************************/
